# Assetto Corsa freejoin server launcher
# by Keijo "Kegetys" Ruotsalainen, http://www.kegetys.fi

import http.server
import socketserver
import urllib.request
import time
import json
import sys
import configparser
import random
import subprocess
import os
import threading
import socket
import struct
import stat

RESERVE_TIME = 15 # How long slot is reserved for a joining player (time for player to connect)
LOCAL_TIMEOUT = 3 # Timeout in seconds for connecting to AC HTTP server
SERVER_TAG = "[FREEJOIN]" # Tag used to detect freejoin servers
cycle_tracks = []
kicked_users = []
cars_list = [] # List of currently active cars
slot_users = dict() # Info on all users (not necessarily connected, only latest slot request info)

def print(*args, **kwargs):
    __builtins__.print("[FREEJOIN] ", end="")
    return __builtins__.print(*args, **kwargs)

def get_server_setting(entry, default=None, cat='SERVER'):
    ini = configparser.ConfigParser()
    ini.read("cfg/server_cfg.ini")
    try:
        return ini[cat][entry]
    except:
        if default == None:
            print("ERROR: Could not read server_cfg file!")
            print("Check you are running server.py from correct path.")
            input("Press enter to quit.")            
            sys.exit()
        else:
            return default

def add_freejoin_setting(ini, name, value):
    if 'FREEJOIN' not in ini.sections():
        ini.add_section('FREEJOIN')
    try: 
        x = ini['FREEJOIN'][name]        
    except KeyError:
        ini['FREEJOIN'][name] = str(value)
        
num_entries = 0
def update_inis():
    # Update server settings
    global num_entries
    ini = configparser.ConfigParser()
    ini.optionxform = str
    ini.read("cfg/server_cfg.ini")
    
    # PICKUP_MODE_ENABLED apparently allows name change mid-race
    ini['SERVER']['PICKUP_MODE_ENABLED'] = "1"
    
    # Cars entry is not used
    ini['SERVER']['CARS'] = ""
    
    if force_restart_time > 0:
        # race over packet ruins everything!
        ini['SERVER']['RACE_OVER_TIME'] = str(force_restart_time+100)
        
    
    # Add defaults for freejoin settings
    add_freejoin_setting(ini, 'RESTART_AFTER_RACE_SECONDS', 80)
    add_freejoin_setting(ini, 'ALLOW_RACE_JOIN', 1)
    add_freejoin_setting(ini, 'ADMIN_PASSWORD', "")
    add_freejoin_setting(ini, 'MOTD', "")
    add_freejoin_setting(ini, 'TRACK_CYCLE', "")
    
    if 'BOOK' in ini.sections():
        print("WARNING: Booking session is active, remove 'BOOK' section from server cfg!")
    
    # Add tag to server name
    name = ini['SERVER']['NAME']
    if(SERVER_TAG not in name):
        ini['SERVER']['NAME'] = "%s %s" % (name, SERVER_TAG)

    # Create unique list of cars
    clist = ""
    for car in set(cars_list):
        clist = clist + "%s;"%car
    ini['SERVER']['CARS'] = clist[:-1]
        
    with open("cfg/server_cfg.ini", 'w') as configfile:
        ini.write(configfile, space_around_delimiters=False)
        
    # Add GUIDs and driver names to entry list
    #ini = configparser.ConfigParser()
    #ini.optionxform = str
    #ini.read("cfg/entry_list.ini")
    #added = 0
    #for idx,s in enumerate(ini.sections()):
    #    if(s[0:4] == "CAR_"):
    #        if idx < len(guids_list):
    #            ini.set(s,'GUID', str(guids_list[idx]))
    #            ini.set(s,'DRIVERNAME', '-')    
    #            added = added+1
    #        else:
    #            print("WARNING: No entry slot for client slot %d" % idx)
    #if added == 0:
    #    print("ERROR: No valid entries in entry_list.ini!")
    #    input("Press enter to quit.")
    #    sys.exit()
    #with open("cfg/entry_list.ini", 'w') as configfile:
    #    ini.optionxform = str
    #    ini.write(configfile, space_around_delimiters=False)
    #num_entries = added
    
def entrylist_from_carini(carini):
    # Create entry_list.ini from cars ini file
    global num_entries, cars_list
    try:
        path = "cfg/cars_%s.ini" % carini.strip()
        if not os.path.isfile(path):            
            return False    
        try:
            out = open("cfg/entry_list.ini", "w")
        except:
            print(sys.exc_info())
            return False
         
        idx = 0     
        temp_list = []
        with open(path, "r") as f:
            for line in f:
                try:
                    if idx < len(guids_list):
                        car,skin = line.strip().split(';')
                        #print(car, skin)
                        out.write("[CAR_%d]\nDRIVERNAME=-\nTEAM=none\nMODEL=%s\nSKIN=%s\nGUID=%d\nSPECTATOR_MODE=0\n" % (idx, car, skin,guids_list[idx]))
                        temp_list.append(car)
                        idx = idx + 1
                except:
                    if(len(line)>1):
                        print("Warning: Invalid car definition: <%s>" % line)
        out.close()
        cars_list = temp_list
        num_entries = idx
        return True
    except:
        print("ERROR: Error parsing car ini <%s>", carini.strip())
        print(sys.exc_info())
        return False
        
def ini_session_enable(ini, session, enabled):
    # Enable or disable session in already opened ini, or add with default
    session_disabled = "%s_DISABLED"%session
    if enabled:
        if session_disabled in ini.sections():
            ini[session] = ini[session_disabled]
            ini.remove_section(session_disabled)
        if not session in ini.sections():
            # Session does not exist, add with some defaults
            ini.add_section(session)        
            ini[session]['NAME'] = session.capitalize()
            ini[session]['WAIT_TIME'] = "25" if session=="RACE" else "10"  # Sensible default?        
    if not enabled and session in ini.sections():
        ini[session_disabled] = ini[session]
        ini.remove_section(session)    
        
def ini_set_track(track, practice_len, quali_len, race_laps):
    # Update ini settings for track & session
    ini = configparser.ConfigParser()
    ini.optionxform = str
    ini.read("cfg/server_cfg.ini")
    
    ini['SERVER']['TRACK'] = track
    
    ini_session_enable(ini, "PRACTICE", practice_len>0)
    ini_session_enable(ini, "QUALIFY", quali_len>0)
    ini_session_enable(ini, "RACE", race_laps>0)
    
    if 'PRACTICE' in ini.sections():
        ini['PRACTICE']['TIME'] = str(practice_len)
    if 'QUALIFY' in ini.sections():
        ini['QUALIFY']['TIME'] = str(quali_len)
    if 'RACE' in ini.sections():
        ini['RACE']['LAPS'] = str(race_laps)
    
    with open("cfg/server_cfg.ini", 'w') as configfile:
        ini.write(configfile, space_around_delimiters=False)    

def do_request(param):
    # Make request to local server
    url = "http://localhost:%d/%s"%(PORT,param)
    req = urllib.request.Request(url=url,data=b'None',headers={'User-Agent':'Assetto Corsa Launcher'})
    return urllib.request.urlopen(req).read().decode("UTF-8")
    
def validate_guid(guid):
    # Check if GUID is valid
    try:
        str = do_request("CANJOIN|%d"%guid)
        if str == "YES":
            return True
        else:
            return False
    except urllib.error.URLError:
        print("Could not connect to local AC server!")
        return False

def check_server():
    # Check if server is running
    try:
        str = do_request("JSON|0")
        if len(json.loads(str)['Cars']) > 0:
            return True
        else:
            return False
    except:
        return False    
            
def server_in_race():
    # Check if server is in race mode
    try:
        str = do_request("ENTRY")
        return "Session: RACE" in str        
    except urllib.error.URLError:
        return False       
            
def slot_free(slot):
    # Check from server if slot is free
    try:        
        str = do_request("JSON|0")
        if json.loads(str)['Cars'][slot]['IsConnected']:
            print("Slot %d not free!" % slot)
            return False
        else:
            return True
    except urllib.error.URLError:
        print("Could not connect to local AC server!")
        return False

def guid_free(guid):
    # Check if guid is free
    if not guid in reserved_guids:
        return True
    if(reserved_guids[guid] > time.time()):
        print("GUID <%s> still reserved" % (guid))        
        return False
    return True
    
def reserve_guid(guid):
    # Reserve guid for player to give him time to join
    reserved_guids[guid] = time.time()+RESERVE_TIME    
    print("GUID <%s> reserved for %d seconds" % (guid, RESERVE_TIME))
    
def slot_get_id(name):
    # Return index of slot with username
    global slot_users
    for slot in slot_users:
        if slot_users[slot][2].lower() == name.lower():
            return int(slot)
    raise ValueError

def slot_get_info(id):
    # Return information of slot
    global slot_users
    try:
        return slot_users[int(id)]
    except:
        raise
        
def slot_get_udp_port(id):
    # Return udp client port with ID
    global slot_users, connected_clients
    try:
        # Find slot user from connected clients
        for client in connected_clients:
            if client[0] == slot_users[int(id)][0]:
                # IP address match
                return int(client[1]) # UDP port
        raise ValueError
    except:
        raise      
    
def validate_track(track):
    # Check if track exists (return string)    
    if os.path.isdir("content/tracks/%s"%track):
        return ""
    else:
        return " !TRACK?"

def validate_car(car):
    # Check if car definition exists (return string)    
    if os.path.isfile("cfg/cars_%s.ini"%car.strip()):
        return ""
    else:
        return " !CAR?"
        
    
connected_clients = [] # All connected  clients
def new_client(ip,port):
    # New client connecting
    # TODO: how to detect a disconnected client?
    global connected_clients
    for idx,c in enumerate(connected_clients):
        if c[0] == ip:
            print("Updated client port %s port %s -> %s" % (ip,c[1],port))
            connected_clients[idx] = [ip,port]
            return
    print("New client (%s,%s)" % (ip,port))
    connected_clients.append([ip,port])

def clear_clients():
    global connected_clients, kicked_users
    connected_clients = []
    kicked_users = []
    
def chat_broadcast(msg):
    # Send chat message to all users
    global slot_users
    sent_list = []
    for user in slot_users:
        target_name = slot_users[user][2]
        if not target_name in sent_list:
            chat_message(target_name, msg)
            sent_list.append(target_name)
    

def chat_message(user, msg):
    # Send chat message to user
    try:
        id = slot_get_id(user)
        ip = slot_get_info(id)[0]
        port = slot_get_udp_port(id)
        fmsg = "S>"+msg
        data = struct.pack("BBBBB", 0xfa, 0x06, 0x47, id, len(fmsg))+bytes(fmsg, 'utf-8')
        try:    
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)  
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)                    
            s.bind(('0.0.0.0', int(UDP_PORT)))
            try:
                s.sendto(data, (ip, port))
            except:
                print("UDP send failed! <%s:%d>" % (ip,int(port)));
                print(sys.exc_info())
                raise
            s.close()
        except:
            print("Socket creation failed!");        
            print(sys.exc_info())
            raise
    except:
        print("WARNING: Could not send chat message to <%s> (user disconnected?)" % user)
        #print(sys.exc_info())
    
def kick_user(id, kicker):
    global kicked_users
    try:
        ip = slot_get_info(id)[0]
        port = slot_get_udp_port(id)
        kicked_users.append(ip)
        if send_to_lobby(ip, port):
            chat_message(kicker, "User kicked (maybe)")
        else:
            chat_message(kicker, "Could not kick user :(")
    except:
        print("Kick error")
        print(sys.exc_info())
            
def send_to_lobby(ip, port):
    # Send client to lobby (magic! :D)                      
    #data = bytes.fromhex("4b01"*512)
    global num_entries
    #data = bytes.fromhex("fa09680100")
    try:    
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)        
        s.bind(('0.0.0.0', int(UDP_PORT)))
        try:
            for i in range(0,num_entries):
                data = bytes.fromhex("fa0968%02x00" %i)
                s.sendto(data, (ip, port))
            #s.sendto(data, (ip, port))
            #time.sleep(0.02)
            #s.sendto(data, (ip, port))
        except:
            print("UDP send failed!");
            print(sys.exc_info())
        s.close()
        return True
    except:
        print("Socket creation failed!");
        print(sys.exc_info())
        return False
    
def send_all_to_lobby():
    # Send all clients to lobby
    for ipp in connected_clients:
        print("Sending client (%s,%s) to lobby" % (ipp[0], ipp[1]));
        try:
            send_to_lobby(ipp[0], int(ipp[1]))
            time.sleep(0.02)
        except ValueError:
            print("Invalid IP or port!");    
    
admins = []
def clear_admins():
    global admins
    admins = []
def is_admin(name):
    return name.lower() in admins
def add_admin(name):
    global admins
    if name.lower() not in admins:
        admins.append(name.lower())
    
MOTDtime = 0
MOTD = ""
def get_motd():
    # Read MOTD once every 30 seconds
    global MOTD, MOTDtime
    try:
        if motd_file == "":
            return ""
        if time.time()-MOTDtime > 0:
            with open(motd_file, 'r') as f:
                MOTD = f.read()
            MOTDtime = time.time()+30
        return MOTD
    except:
        return ""
     
def handle_motd(request, rHost, rPort):
    # Handler for message of the day
    # request contains user name and guid also here
    return get_motd()

def handle_reqjoin(request, rHost, rPort):
    # Handler for "REQJOIN"
    global slot_users
    req = request.split('|')
    print("reqjoin %d" % len(req))
    if(len(req) < 5):
        return ""
    try:
        rPassword = req[1] # Requested password
        rId = int(req[2])  # Index of wanted slot
        rRealGuid = req[3] # Real GUID of player
        rRealName = req[4] # Player name
        isAdmin = False    # Player connected with admin password
        
        error = None
        if(rPassword == PASSWORD_ADMIN and len(PASSWORD_ADMIN)>0):
            isAdmin = True
        elif(rPassword != PASSWORD and len(PASSWORD)>0):
            error = "Invalid password"        
        elif(rId<0 or rId >= len(guids_list)):
            error = "Invalid car ID"
        elif(not allow_race_join and server_in_race()):
            error = "Mid-race join disabled"
        elif(rRealName.lower() == "3dmgame"):
            error = "Error 0x574152455A42414E"
        elif(rHost in kicked_users):
            error = "Not welcome"
        elif is_admin(rRealName) and not isAdmin:
            # Player previously connected with admin password
            # or another user is trying to fake admin
            error = "Connect with admin password"
        
        if error == None:
            # Valid request, check GUID
            print("Slot request: %d from %s:%d (%s:%s)" % (rId, rHost, rPort, rRealName, rRealGuid))
            valid_guid = str(guids_list[rId])
            if not guid_free(valid_guid):
                error = "Slot reserved, refresh player list"
            elif not slot_free(rId):
                error = "Slot occupied, refresh player list"
            if error == None:
                # Send GUID to player
                reserve_guid(valid_guid) # GUID is reserved for a while
                slot_users[rId] = [rHost, rPort, rRealName, rRealGuid]
                addPass = "|"
                if isAdmin:
                    add_admin(rRealName)
                    print("User %s is administrator" % rRealName)
                    addPass = "|%s"%PASSWORD # Send join password for admin
                return "OK|%s%s"%(valid_guid,addPass)
                
        # Wrong password or unjoinable slot
        print("Invalid request for slot %d, error: <%s> from %s:%d (%s:%s)" % (int(rId), str(error), rHost, rPort, rRealName, rRealGuid))
        return "NO|%s"%error
    except ValueError:            
        return ""
    
class rhandler(http.server.BaseHTTPRequestHandler):
    def log_message(self, format, *args):
        pass
        
    def do_GET(self):
        # Incoming HTTP request handler
        try:
            path = urllib.request.url2pathname(self.path)
            request = path.split('|',1)[0][1:]
            response = ""
            if request == "REQJOIN":
                response = handle_reqjoin(path, self.client_address[0], self.client_address[1])
            elif request == "MOTD":
                response = handle_motd(path, self.client_address[0], self.client_address[1])
            self.send_response(200)
            self.send_header('Content-type','text-html')
            self.end_headers()
            self.wfile.write(response.encode("utf-8"))            
        except:
            pass

def skip_restart():
    # Abort an ongoing delay restart
    global restart_id
    restart_id = 0
   
want_restart = False   
restart_id = 0
def server_restart_thread(process, rtime, local_restart):
    global want_restart, restart_id
    remain = rtime-1
    while remain > 0:
        time.sleep(1)
        remain = remain - 1
        for warn in [10,30,60,120]:
            if remain<warn and remain+1>=warn:
                chat_broadcast("Next race in %d seconds" % warn)
        if local_restart != restart_id:
            print("Restart aborted")
            # Restart aborted
            return
    try:
        if process and local_restart == restart_id:
            want_restart = True
            chat_broadcast("Next race!")
            time.sleep(0.1)            
            process.terminate()
            check_server() # Ugly hack: Make server output something to stdout so stout reader wont block
        else:
            print("Restart aborted")
    except:
        print("Could not restart process!");
    
def trigger_restart(process, rtime):
    # Stat bg thread to restart server
    global restart_id
    restart_id = random.randint(1, 1000000)
    thread = threading.Thread(target=server_restart_thread, args=(process, rtime, restart_id)).start()
 
def handle_chat(process, line):
    global want_restart, restart_id
    user = ""
    msg = ""
    try:
        a = line.split(':',1)
        user = a[0].split('[',1)[1].split(']',1)[0].strip()
        msg = a[1].strip()
        print("<%s:%s>" %(user,msg))
    except:
        print("Invalid chat message: <%s>" % line)
        return        

    # Handle admin commands
    if is_admin(user) and msg[0] == '!':     
        cmd,params = (msg+" ").split(" ",1)
        cmd = cmd.strip()
        params = params.strip()
        if cmd == "!ping":
            print("ADMIN: hello world")
            chat_message(user, "pong!")       
        if cmd == "!tracks":
            print("ADMIN: Set track cycle")            
            if parse_track_cycle(params, False):
                chat_message(user, "Track cycle:")
                for c in cycle_tracks:
                    chat_message(user, " %s %s: p%dm q%dm r%dl %s%s" % (c[0],c[1],c[2],c[3],c[4],validate_track(c[0]),validate_car(c[1])))
            else:
                chat_message(user, "Invalid parameters!")
        if cmd == "!next":
            print("ADMIN: Next race")
            chat_broadcast("Admin: next race")
            want_restart = True
            send_all_to_lobby()
            clear_admins()
            clear_clients()            
            process.terminate()
            check_server() # Ugly hack: Make server output something to stdout so stout reader wont block
        if cmd == "!users":
            print("ADMIN: User list")
            for suser in slot_users:
                print(suser,slot_users[suser])
                chat_message(user, "%d: %s, %s, %s" % (int(suser), slot_users[suser][2], slot_users[suser][0], slot_users[suser][3]))
        if cmd == "!skip":
            if restart_id > 0:
                chat_broadcast("Admin: Next race cancelled")
            skip_restart()
        if cmd == "!kick":
            try:
                id = int(params)
                print("ADMIN: Kicking user ID %d" % id)
                kick_user(id, user)
            except:
                chat_message(user, "Invalid parameters")
                print(sys.exc_info())
        if cmd == "!help":
            chat_message(user, "Commands:")
            chat_message(user, " !users: List users")
            chat_message(user, " !kick <id>: Kick user number id")
            chat_message(user, " !skip: After race end, cancel auto-restart")
            chat_message(user, " !next: Go to next race")
            chat_message(user, " !tracks: List track cycle")
            chat_message(user, " !tracks <c>: Set track cycle ex: imola,default,1,2,3;drift,bmw,2,2,3")
            chat_message(user, " !ping: pong!")

def process_server_output(line, process):
    # Proces stdout output from AC server
    if("NEW CONNECTION from " in line):
        # message: "NEW CONNECTION from 89.166.96.115:50132" - new client is connecting
        # Get IP and port
        ipp = line[len("NEW CONNECTION from "):].strip(' \r\n')
        ip,port = ipp.split(':')
        new_client(ip,port)
        
    if("RACE OVER DETECTED!" in line):
        # Race has ended        
        if force_restart_time > 0:
            print("Race end, terminating server in %d seconds" % force_restart_time)
            chat_broadcast("Race finished, next race in %d seconds" % force_restart_time)
            trigger_restart(process, force_restart_time)
            
    if(line[0:5] == "CHAT "):
        handle_chat(process, line)
        
       
def server_handler_thread():
    # Thread for handling the acserver
    global want_restart
    print("Starting ACserver")
    
    serverProcess = None
    doRun = True    
    try:
        while doRun:
            try:
                executable = "./acServer.exe"
                if os.name == "posix":
                    executable = "./run_server"
                    if sys.maxsize > 2**32:
                        #executable = "./run_server_x64"
                        print("Warning: Running x64 server despite x64 architexture");                    
                    os.chmod(executable, os.stat(executable).st_mode | stat.S_IEXEC) # Make sure binary is executable                  
                serverProcess = subprocess.Popen([executable],stdout=subprocess.PIPE)
            except:
                print("ERROR: Could not start server.")
                print("Check server.py is ran from correct path.")
                sys.exit()
            result = serverProcess.poll()
            #while not serverProcess.poll():
            while not result:
                # Read server output
                line = serverProcess.stdout.readline()
                str = ""
                try:
                    str = line.decode("UTF-8")
                except:
                    # Chat messages are not utf8
                    print("ERROR: outout decode error")
                process_server_output(str, serverProcess)
                __builtins__.print(str, end="")
                result = serverProcess.poll()
            print("Server process terminated with signal %d" % result)
            #if result == 2 and not want_restart:
            if not want_restart:
                # ctrl-c
                print("Terminating handler")                
                send_all_to_lobby()
                clear_clients()
                doRun = False
            else:
                print("Restarting server")
                send_all_to_lobby()
                cycle_track()
                clear_admins()
                clear_clients()
                time.sleep(4)
                want_restart = False
    except KeyboardInterrupt:
        if serverProcess:
            serverProcess.terminate()
          
def parse_track_cycle(cycle,critical=True):
    global cycle_tracks, track_index
    try:
        new_cycle = []
        if(len(cycle.strip())>1):
            for entry in cycle.split(';'):
                if(len(entry.strip()) > 1):
                    data = entry.split(',',4)
                    data[0] = str(data[0])
                    data[1] = str(data[1])
                    data[2] = int(data[2])
                    data[3] = int(data[3])
                    data[4] = int(data[4])
                    new_cycle.append(data)
            cycle_tracks = new_cycle
            track_index = 0
        return True
    except:
        print("ERROR: Invalid track cycle defined!")
        print(sys.exc_info())
        if critical:
            input("Press enter to quit.")
            sys.exit()
        return False

track_index = 0        
def cycle_track():
    # Do track cycling based on TRACK_CYCLE setting, update track & session parameters
    global cycle_tracks, track_index
    if(len(cycle_tracks) == 0):
        return
    track = cycle_tracks[track_index%len(cycle_tracks)]
    ini_set_track(track[0], track[2], track[3], track[4])
    print("Track changed to %s with p%dmin, q%dmin, r%dlaps" % (track[0], track[2], track[3], track[4]))
    if not entrylist_from_carini(track[1]):
        print("WARNING: Could not parse car ini!")
    else:
        update_inis()
    track_index = track_index + 1
    
# Read settings from server cfg
PORT = int(get_server_setting('HTTP_PORT'))
UDP_PORT = int(get_server_setting('PORT'))
PASSWORD = get_server_setting('PASSWORD')
num_clients = int(get_server_setting('max_clients'))
allow_race_join = int(get_server_setting('ALLOW_RACE_JOIN', "1", "FREEJOIN"))>0
PASSWORD_ADMIN = get_server_setting('ADMIN_PASSWORD', "", "FREEJOIN")
motd_file = get_server_setting('MOTD', "", "FREEJOIN")
force_restart_time = int(get_server_setting('RESTART_AFTER_RACE_SECONDS', "80", "FREEJOIN"))
parse_track_cycle(get_server_setting('TRACK_CYCLE', "", "FREEJOIN"))

# Create random GUIDs for players
guids_list = random.sample(range(1000, 100000000), num_clients)
reserved_guids = dict()

# Update config files
if not entrylist_from_carini("default"):
    print("ERROR: cfg/cars_default.ini not found!")
    input("Press enter to quit.")
    sys.exit()    
update_inis()

print("Server name <%s>" % get_server_setting('NAME'))
print("Clients: %d, HTTP port: %d, passsword: <%s>" % (num_clients, PORT, PASSWORD));

if len(cycle_tracks)>0:
    print("Track cycle:")
    for c in cycle_tracks:
        print(" <%s,%s> practice: %dmin, qualifying: %dmin, race: %dlaps %s%s" % (c[0],c[1],c[2],c[3],c[4],validate_track(c[0]),validate_car(c[1])))
else:
    print("No track cycle")

cycle_track() # Select first track from track list
        
# Check server is not running
if check_server():
    print("ERROR: Server is already running!")
    sys.exit()

# Run the server, wait for it to init
thread = threading.Thread(target=server_handler_thread).start()
time.sleep(4)
    
# Check that the local server is running
while not check_server():
    print("WARNING: Could not connect to AC server!")
    time.sleep(2)
print("Connected to local AC server")
   
# Validate guids list from server
for i in range(0, num_entries):    
    guid = guids_list[i]
    if not validate_guid(guid):
        print("ERROR: GUID <%s> is not valid!" % guid)
        sys.exit()
print("Validated %d GUIDs" % num_entries)
    
# Run HTTP server
serve_port = PORT+1
class fjTcpServer(socketserver.TCPServer):
    allow_reuse_address = True # Avoid bind problems after terminating server with ctrl-c
httpd = fjTcpServer(("", serve_port), rhandler)
print("Serving at port", serve_port)
try:
    httpd.serve_forever()
except KeyboardInterrupt:
    print("Terminated HTTP server")